/*
 * BQ4SURFIT
 *
 *      This subroutine takes the (x,y,z) data arrays and
 *      the mean data values (xmean,ymean,zmean) fits a
 *      biquartic surface function with 15 coefficients
 *      which are returned in (cvec) vector.
 *
 *  Author:     Paul J. Besl    (4-18-85)
 *
 *  Revised:    Arnold. H. Chiu (7-26-88)
 *  
 *  Call:           bq4surfit(x_array,y_array,z_array,num_pixels,
 *                            x_mean,y_mean,error_total,x_size,y_size,
 *                            cvec,opt,ovec)
 *  Modules used:   linlsq(),   matcpyd(),  bq4convert(),
 *
 */

#define ARSIZE 15

bq4surfit(xray,yray,zray,npixels,xmean,ymean,errtol,nx,ny,cvec,opt,ovec)
float   *xray,*yray,*zray,xmean,ymean,*errtol; 
int     npixels,nx,ny,opt;
double  cvec[],ovec[];
{
    unsigned char *malloc();
    float *xptr,*yptr,*zptr;
    double *phiray,x2,x3,y2,y3,*zdblvec,error,x42,y42,x40,y40;
    double xval2,yval2,xval3,yval3,x2val,y2val,nx2,ny2;
    register double *phiptr,xval,yval,*zdblptr;
    register int i;
    int rank,arsize = ARSIZE;

    if( npixels < arsize )
    {
        printf("=> %d Data Points needed for Surface Fit\n",arsize);
        exit(1);
    }

    /* Allocate memory for npixels x arsize float array. */

    phiray  = (double *)malloc(8*arsize*npixels);
    zdblvec = (double *)malloc(8*npixels);

    /* Compute Quadratic and Cubic Offset Values */

    nx2 = nx*nx;                        ny2 = ny*ny;
    x2 = ( 1.0 - nx2 ) / 12.0;          y2 = ( 1.0 - ny2 ) / 12.0;
    x3 = ( 7.0 - 3.0*nx2 ) / 20.0;      y3 = ( 7.0 - 3.0*ny2 ) / 20.0;
    x42 = ( 13.0 - 3.0*nx2 ) / 14.0;    y42 = ( 13.0 - 3.0*ny2 ) / 14.0;
    x40 = ( 3.0*(nx2 - 1.0)*(nx2 - 9.0) ) / 560.0;
    y40 = ( 3.0*(ny2 - 1.0)*(ny2 - 9.0) ) / 560.0;

    /* Loop over all data points filling in PHI array */

    xptr = xray; yptr = yray; zptr = zray;
    phiptr = phiray;  zdblptr = zdblvec;
    for(i=0;i<npixels;++i)
    {
        xval = (*xptr++) - xmean;  yval = (*yptr++) - ymean;
        *zdblptr++ = *zptr++;
        xval2 = xval*xval;         yval2 = yval*yval;
        x2val = xval2 + x2;        y2val = yval2 + y2;
        xval3 = (xval2+x3)*xval;   yval3 = (yval2+y3)*yval;
        *phiptr++ = 1;             /* 0 */
        *phiptr++ = xval;          /* 1 */
        *phiptr++ = yval;          /* 2 */
        *phiptr++ = xval*yval;     /* 3 */
        *phiptr++ = x2val;         /* 4 */
        *phiptr++ = y2val;         /* 5 */
        *phiptr++ = x2val*yval;    /* 6 */
        *phiptr++ = xval*y2val;    /* 7 */
        *phiptr++ = xval3;         /* 8 */
        *phiptr++ = yval3;         /* 9 */
        *phiptr++ = xval3*yval;    /* 10 */
        *phiptr++ = x2val*y2val;   /* 11 */
        *phiptr++ = xval*yval3;    /* 12 */
        *phiptr++ = xval2*xval2 + x42*xval2 + x40;  /* 13 */
        *phiptr++ = yval2*yval2 + y42*yval2 + y40;  /* 14 */
    }

    /* Compute least squares solution */

    rank = linlsq(cvec,phiray,zdblvec,npixels,arsize,opt,&error);
    matcpyd(ovec,cvec,arsize,1);
    bq4convert(xmean,ymean,nx,ny,cvec);
    printf("[Biquartic_Fit(4)] RMS_Error= %g\n",error);
    free( zdblvec );
    free( phiray );
    *errtol = error;
}
