/*
 *  LINLSQ
 *
 *      This routine is a linear least squares solver.
 *      [A][x] = [b]        { [x]=(ncols x 1) } 
 *      { [A]=(nrows x ncols) [b]=(nrows x 1) }
 *      This subroutine solves Ax = b [nrows x ncols].
 *      Opt = 0 --> Use C - subroutines for matrix operations
 *      Opt = 1 --> Use F77 - subroutine for lin. lst. sqrs.
 *      Opt = 2 --> Use C with Inverse Quality Check
 *      Returns Rank of ATA matrix.
 *
 *  Author:     Paul J. Besl    (4-16-85)
 *
 *  Revised:    Arnold. H. Chiu (7-26-88)
 *  
 *  Call:           linlsq(xvec,aray,bvec,nrows,ncols,opt,error)
 *  Modules used:   mattpsd(),  dllsq_() {fortran routine},
 *                  ifree(),    dfree(),    dmalloc(),  matmuld(),
 *                  matatad(),  matcpyd(),  matsinvd(),  matnrmd()
 *
 */

/* pjf 4/90 */
#include <stdio.h>
/* pjf 4/90 */

/* stop listing <math.h> */
/* #nolist */
#include <math.h>
/* #list */

int linlsq(xvec,aray,bvec,nrows,ncols,opt,error)
double  xvec[],aray[],bvec[],*error;
int     nrows,ncols,opt;
{
    int         *imalloc();
    int         matsinvd(),tsecs,tstop();
/* pjf 4/90 */
    /*std_$call   dllsq(); */
/* pjf 4/90 */
    double      *ataray,*atray,*atbvec,*aux;
    double      nrm1,nrm2,nrm3,*dmalloc();
    int         *ipiv,norm,rank,ier,one = 1;
    float       eps = 1.0e-12;

    /* Allocate memory for ncols x nrows float array. */

    if( opt == 1 )
    {
/* pjf 4/90 */
      fprintf(stderr,"F77 fit not available.\n");
/* pjf 4/90 */

        /* Use F77 subroutine */
/* pjf 4/90 */
/* don't.
        atray  = dmalloc(nrows*ncols);
        aux    = dmalloc(2*ncols);
        ipiv   = imalloc(ncols);
        mattpsd(atray,aray,nrows,ncols);
        dllsq(atray,bvec,&nrows,&ncols,&one,xvec,ipiv,&eps,&ier,aux);
        if( ier == 0 )
            rank = ncols;
        else
            rank = ier;
        *error = aux[0];
        ifree( ipiv );
        dfree( aux );
        dfree( atray );
*/
/* pjf 4/90 */
    }
    else /* This is Good Math, but Bad Numerical Analysis (but It Works!) */
    {
        /* Allocate Memory and Compute Error Norm Quantity */

        atbvec = dmalloc(ncols*(ncols+1));
        ataray = atbvec + ncols;
        if( opt == 2 ) atray  = dmalloc(ncols*ncols);
        matmuld(&nrm1,bvec,bvec,1,nrows,1);      /* nrm1 = [b]T[b] */

        /* Compute [A]T[A] Matrix for inversion and [A]T[b] */

        matatad(ataray,aray,nrows,ncols);        /* ata = [A]T[A] */
        matmuld(atbvec,bvec,aray,1,nrows,ncols); /* atb = [A]T[b] */
        if( opt == 2 )
            matcpyd(aray,ataray,ncols,ncols);
        /*tsecs = tstart();*/

       /* Invert [A]T[A] matrix of Normal Equations [A]T[A][x] = [A]T[b] */

        rank = matsinvd(ataray,ncols);
        /*tstop(tsecs,"Sym_Mat_Inv_Time");*/

        /* Compute Solution of Normal Equations [x] = inverse([A]T[A])[A][b] */

        matmuld(xvec,ataray,atbvec,ncols,ncols,1); /* x = [ata]-1[atb] */
        matmuld(&nrm2,xvec,atbvec,1,ncols,1);
        *error = nrm1 - nrm2;

        /* Check Pseudo-Inverse Quality if Opted For */ 

        if( opt == 2 )  /* aray contains original copy of ataray */
        {
            matmuld(atray,aray,ataray,ncols,ncols,ncols);
            matnrmd(atray,ncols,&nrm1,&nrm2,&nrm3);
            norm = (int)(nrm3*nrm3 + 0.5);
            if( norm > ncols )
            { 
                printf("(1) Identity Matrix Norm= %d\n",norm);
                exit(1);
            }
            else
            {
                rank = norm;
                /*printf("(1) Inverse Check OK\n");*/
            }
            matmuld(atray,ataray,aray,ncols,ncols,ncols);
            matnrmd(atray,ncols,&nrm1,&nrm2,&nrm3);
            norm = (int)(nrm3*nrm3 + 0.5);
            if( norm > ncols ) 
            {
                printf("(2) Identity Matrix Norm= %d\n",norm);
                exit(1);
            }
            else
            {
                rank = norm;
                /*printf("(2) Inverse Check OK\n");*/
            }
        }
        if( opt == 2 )
            dfree( atray );
        dfree( atbvec ); 
    }
    *error /= (double)nrows;
    *error = sqrt( *error );
    /*printf("System_Rank= %d\n",rank);*/
    /*printf("RMS_Error= %g\n",*error);*/
    return( rank );
}
